@@ -30,13 +30,17 @@ from ema_pytorch import EMA
 
 from accelerate import Accelerator, DistributedType, DistributedDataParallelKwargs
 
+from fsspec.implementations.local import LocalFileSystem
+
 # helper functions
 
 def exists(val):
     return val is not None
 
 def default(val, d):
-    return val if exists(val) else d
+    if exists(val):
+        return val
+    return d() if callable(d) else d
 
 def cast_tuple(val, length = 1):
     if isinstance(val, list):
@@ -211,6 +215,7 @@ class ImagenTrainer(nn.Module):
         split_random_seed = 42,
         checkpoint_path = None,
         checkpoint_every = None,
+        checkpoint_fs = None,
         max_checkpoints_keep = 20,
         **kwargs
     ):
@@ -218,10 +223,14 @@ class ImagenTrainer(nn.Module):
         assert not ImagenTrainer.locked, 'ImagenTrainer can only be initialized once per process - for the sake of distributed training, you will now have to create a separate script to train each unet (or a script that accepts unet number as an argument)'
         assert exists(imagen) ^ exists(imagen_checkpoint_path), 'either imagen instance is passed into the trainer, or a checkpoint path that contains the imagen config'
 
-        if exists(imagen_checkpoint_path):
-            checkpoint_path = Path(imagen_checkpoint_path)
-            assert checkpoint_path.exists()
-            loaded = torch.load(str(imagen_checkpoint_path))
+        # determine filesystem, using fsspec, for saving to local filesystem or cloud
+
+        fs = default(checkpoint_fs, lambda: LocalFileSystem())
+        self.fs = fs
+
+        if self.fs.exists(imagen_checkpoint_path):
+            with self.fs.open(imagen_checkpoint_path) as f:
+                loaded = torch.load(imagen_checkpoint_path)
 
         assert isinstance(imagen, (Imagen, ElucidatedImagen))
         ema_kwargs, kwargs = groupby_prefix_and_trim('ema_', kwargs)
@@ -336,10 +345,11 @@ class ImagenTrainer(nn.Module):
         self.checkpoint_every = checkpoint_every
         self.max_checkpoints_keep = max_checkpoints_keep
 
-        if exists(checkpoint_path) and self.is_local_main:
-            self.checkpoint_path = Path(checkpoint_path)
-            self.checkpoint_path.mkdir(exist_ok = True, parents = True)
+        self.can_checkpoint = self.is_local_main if isinstance(checkpoint_fs, LocalFileSystem) else self.is_main
 
+        if exists(checkpoint_path) and self.can_checkpoint:
+            if not fs.exists(checkpoint_path):
+                self.fs.mkdir(checkpoint_path)
             self.load_from_checkpoint_folder()
 
         # only allowing training for unet
@@ -553,40 +563,37 @@ class ImagenTrainer(nn.Module):
 
     @property
     def all_checkpoints_sorted(self):
-        checkpoints = [*self.checkpoint_path.glob('*.pt')]
+        glob_pattern = os.path.join(self.checkpoint_path, '*.pt')
+        checkpoints = self.fs.glob(glob_pattern)
         sorted_checkpoints = sorted(checkpoints, key = lambda x: int(str(x).split('.')[-2]), reverse = True)
         return sorted_checkpoints
 
     def load_from_checkpoint_folder(self, last_total_steps = -1):
         if last_total_steps != -1:
-            filepath = str(self.checkpoint_path / f'checkpoint.{last_total_steps}.pt')
-            self.load(str(filepath))
+            filepath = os.path.join(self.checkpoint_path, f'checkpoint.{last_total_steps}.pt')
+            self.load(filepath)
             return
 
         sorted_checkpoints = self.all_checkpoints_sorted
 
         if len(sorted_checkpoints) == 0:
-            self.print(f'no checkpoints found to load from at {str(self.checkpoint_path)}')
+            self.print(f'no checkpoints found to load from at {self.checkpoint_path}')
             return
 
         last_checkpoint = sorted_checkpoints[0]
-        self.load(str(last_checkpoint))
-
-        self.print(f'loading checkpoint from {str(last_checkpoint)}')
+        self.load(last_checkpoint)
 
     def save_to_checkpoint_folder(self):
         self.accelerator.wait_for_everyone()
 
-        if not self.is_local_main:
+        if not self.can_checkpoint:
             return
 
         total_steps = int(self.steps.sum().item())
-        filepath = self.checkpoint_path / f'checkpoint.{total_steps}.pt'
+        filepath = os.path.join(self.checkpoint_path, f'checkpoint.{total_steps}.pt')
 
         self.save(filepath)
 
-        self.print(f'saved checkpoint to {str(filepath)}')
-
         if self.max_checkpoints_keep <= 0:
             return
 
@@ -594,19 +601,24 @@ class ImagenTrainer(nn.Module):
         checkpoints_to_discard = sorted_checkpoints[self.max_checkpoints_keep:]
 
         for checkpoint in checkpoints_to_discard:
-            os.remove(str(checkpoint))
+            self.fs.rm(checkpoint)
 
     # saving and loading functions
 
     def save(self, path, overwrite = True, **kwargs):
         self.accelerator.wait_for_everyone()
 
-        if not self.is_local_main:
+        if not self.can_checkpoint:
             return
 
-        path = Path(path)
-        assert not (path.exists() and not overwrite)
-        path.parent.mkdir(parents = True, exist_ok = True)
+        fs = self.fs
+
+        assert not (fs.exists(path) and not overwrite)
+
+        dirname = os.path.dirname(path)
+
+        if not fs.exists(dirname):
+            fs.mkdir(dirname)
 
         self.reset_ema_unets_all_one_device()
 
@@ -652,22 +664,26 @@ class ImagenTrainer(nn.Module):
 
         #save to path
 
-        torch.save(save_obj, str(path))
-        self.print(f'checkpoint saved to {str(path)}')
+        with fs.open(path, 'wb') as f:
+            torch.save(save_obj, f)
+
+        self.print(f'checkpoint saved to {path}')
 
     def load(self, path, only_model = False, strict = True, noop_if_not_exist = False):
-        path = Path(path)
+        fs = self.fs
 
-        if noop_if_not_exist and not path.exists():
+        if noop_if_not_exist and not fs.exists(path):
             self.print(f'trainer checkpoint not found at {str(path)}')
             return
 
-        assert path.exists(), f'{str(path)} does not exist'
+        assert fs.exists(path), f'{path} does not exist'
 
         self.reset_ema_unets_all_one_device()
 
         # to avoid extra GPU memory usage in main process when using Accelerate
-        loaded_obj = torch.load(str(path), map_location='cpu')
+
+        with fs.open(path) as f:
+            loaded_obj = torch.load(f, map_location='cpu')
 
         if version.parse(__version__) != version.parse(loaded_obj['version']):
             self.print(f'loading saved imagen at version {loaded_obj["version"]}, but current package version is {__version__}')
@@ -705,7 +721,7 @@ class ImagenTrainer(nn.Module):
             assert 'ema' in loaded_obj
             self.ema_unets.load_state_dict(loaded_obj['ema'], strict = strict)
 
-        self.print(f'checkpoint loaded from {str(path)}')
+        self.print(f'checkpoint loaded from {path}')
         return loaded_obj
 
     # managing ema unets and their devices
